#ifndef _JACUTILS_H_
#define _JACUTILS_H_
/*
  ----------------------------------------------------------------- 
  File jacutils.h of ADOL-C version 1.8.0           as of Dec/03/98
  -----------------------------------------------------------------
  This file containts declarations of jacobian utilities. 

  Last changes:
    981203 olvo:    untransposing reverse
    981130 olvo:    includes changed
    981126 olvo:    last check (p's & q's)
    981125 christo: changed block_pattern() parameter list 
    981118 christo: changed block_pattern() parameter list
    981101 christo: changed block_pattern() parameter values
  
  -----------------------------------------------------------------
*/

/****************************************************************************/
/*                                                                   MAKROS */

/* Max. number of unsigned ints to store the seed / jacobian matrix strips.  
   Reduce this value to x if you system happens to run out of memory. 
   x < 10 makes no sense. x = 50 or 100 is better
   x stays for ( x * sizeof(unsigned int) * 8 ) (block) variables at once   */
#define PQ_STRIPMINE_MAX 30


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/*                                                      PUBLIC EXPORTS ONLY */

#ifdef __cplusplus
/****************************************************************************/
/****************************************************************************/
/*                                                            No C++ THINGS */


/****************************************************************************/
/****************************************************************************/
/*                                                         Now the C THINGS */
extern "C" {
#endif

/****************************************************************************/
/*                                                    BIT PATTERN UTILITIES */

/*--------------------------------------------------------------------------*/
/* int_forward_tight(tag, m, n, p, x[n], X[n][p], y[m], Y[m][p])            */

int int_forward_tight(short, int, int, int, double*, unsigned int**,
                                            double*, unsigned int**);


/*--------------------------------------------------------------------------*/
/* int_forward_safe(tag, m, n, p, X[n][p], Y[m][p])                         */

int int_forward_safe(short, int, int, int, unsigned int**, unsigned int**);


/*--------------------------------------------------------------------------*/
/* int_reverse_tight(tag, m, n, q, U[q][m], Z[q][n])                        */

int int_reverse_tight(short, int, int, int, unsigned int**, unsigned int**);


/*--------------------------------------------------------------------------*/
/* int_reverse_safe(tag, m, n, q, U[q][m], Z[q][n])                         */

int int_reverse_safe(short, int, int, int, unsigned int**, unsigned int**);


/****************************************************************************/
/*                                                   JACOBIAN BLOCK PATTERN */

int block_pattern(short, int, int, double*, unsigned short*, unsigned short*,
                                            unsigned short**, int[]);


/****************************************************************************/
/*                                              MEMORY MANAGEMENT UTILITIES */

unsigned short int * myalloc1_ushort(unsigned short int);

unsigned int *  myalloc1_uint(int);
unsigned int ** myalloc2_uint(int, int);


/****************************************************************************/
/*                                                               THAT'S ALL */
#ifdef __cplusplus
}
#endif

#endif
