#define _SPARSEC_CPP_
#define _ADOLC_SRC_
/*
   --------------------------------------------------------------
   File sparseC.C of ADOL-C version 1.8.0         as of Dec/03/98
   --------------------------------------------------------------
   All "Easy To Use" C++ interfaces of SPARSE package

   Last changes:
        981203 olvo: untransposing reverse
        981201 olvo: newly created from interfaces.C

   --------------------------------------------------------------
*/


/****************************************************************************/
/*                                                                 INCLUDES */
#include "dvlparms.h" /* Developers Parameters */
#include "usrparms.h" /* Users Parameters      */
#include "sparse.h"
#include "jacutils.h"

#include <stdio.h>
#include <math.h>
#include <stdlib.h>


/****************************************************************************/
/*                                    Bit pattern propagation; general call */
/*                                                                          */
int forward( short        tag,
             int          m,
             int          n,
             int          p,
             double       *x,
             unsigned int **X,
             double       *y,
             unsigned int **Y,
             short        mode)
/* forward(tag, m, n, p, x[n], X[n][p], y[m], Y[m][p], mode)                */
{
  int rc = -1;
  if (mode == 1) // tight version
    if (x != NULL) 
      rc = int_forward_tight(tag,m,n,p,x,X,y,Y);
    else
    { fprintf(DIAG_OUT,"ADOL-C error:  no basepoint for bit"
                         " pattern forward tight.\n");
        exit(-1);
    }
  else 
    if (mode == 0) // safe version
      rc = int_forward_safe(tag,m,n,p,X,Y);
    else
    { fprintf(DIAG_OUT,"ADOL-C error:  bad mode parameter to bit"
                       " pattern forward.\n");
      exit(-1);
    }
  return (rc);
}


/****************************************************************************/
/*                                    Bit pattern propagation; no basepoint */
/*                                                                          */
int forward( short        tag,
             int          m,
             int          n,
             int          p,
             unsigned int **X,
             unsigned int **Y,
             short        mode)
/* forward(tag, m, n, p, X[n][p], Y[m][p])                                  */
{ 
  if (mode != 0) // not safe
  { fprintf(DIAG_OUT,"ADOL-C error:  bad mode parameter to bit"
                     " pattern forward.\n");
    exit(-1);
  }
  return int_forward_safe(tag,m,n,p,X,Y);
}



/****************************************************************************/
/*                                                                          */
/*         Bit pattern propagation, always tight version                    */
/*         !!! Attention : U and Z with reversed order of indeces           */
/*                                                                          */
int reverse( short        tag, 
             int          m,
	     int          n,
             int          q,
	     unsigned int **U,
	     unsigned int **Z,
             short        mode)
/* reverse(tag, m, n, 0, q, U[q][m], Z[q][n]), d=0 => Z matrix */
{ int rc=-1;

  /* ! use better the tight version, the safe version supports no subscripts*/

  if (mode == 0) // safe version
    rc = int_reverse_safe(tag,m,n,q,U,Z);
  else
    if (mode == 1)
      rc = int_reverse_tight(tag,m,n,q,U,Z);
    else
    { fprintf(DIAG_OUT,"ADOL-C error:  bad mode parameter"
                       " to bit pattern reverse.\n");
      exit(-1);
    }
  return rc;
}


/****************************************************************************/
/*                                                               THAT'S ALL */
#undef _ADOLC_SRC_
#undef _SPARSEC_CPP_
